<?php

namespace Drupal\crowdsec;

use CrowdSec\CapiClient\Storage\StorageInterface;
use CrowdSec\CapiClient\Watcher;
use CrowdSec\RemediationEngine\CacheStorage\AbstractCache;
use CrowdSec\RemediationEngine\CacheStorage\Redis;
use CrowdSec\RemediationEngine\CacheStorage\PhpFiles;
use CrowdSec\RemediationEngine\CapiRemediation;
use CrowdSec\RemediationEngine\Constants;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\File\FileSystem;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Site\Settings;
use Psr\Cache\CacheException;
use Psr\Cache\InvalidArgumentException;

/**
 * Provides the CrowdSec watcher interface.
 */
class Client {

  /**
   * The CrowdSec configuration.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected ImmutableConfig $config;

  /**
   * The storage service.
   *
   * @var \CrowdSec\CapiClient\Storage\StorageInterface
   */
  protected StorageInterface $storage;

  /**
   * The cache for the remediation engine.
   *
   * @var \CrowdSec\RemediationEngine\CacheStorage\AbstractCache|null
   */
  protected ?AbstractCache $cache;

  /**
   * The CrowdSec watcher.
   *
   * @var \CrowdSec\CapiClient\Watcher|null
   */
  protected ?Watcher $watcher;

  /**
   * The CrowdSec remediation engine.
   *
   * @var \CrowdSec\RemediationEngine\CapiRemediation|null
   */
  protected ?CapiRemediation $remediation;

  /**
   * The logger service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected LoggerChannelInterface $logger;

  /**
   * The file system service.
   *
   * @var \Drupal\Core\File\FileSystem
   */
  protected FileSystem $fileSystem;

  /**
   * Constructs the CrowdSec client implementation.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   The config factory service.
   * @param \CrowdSec\CapiClient\Storage\StorageInterface $storage
   *   The storage service.
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger
   *   The logger service.
   * @param \Drupal\Core\File\FileSystem $fileSystem
   *   The file system service.
   */
  public function __construct(ConfigFactoryInterface $configFactory, StorageInterface $storage, LoggerChannelInterface $logger, FileSystem $fileSystem) {
    $this->config = $configFactory->get('crowdsec.settings');
    $this->storage = $storage;
    $this->logger = $logger;
    $this->fileSystem = $fileSystem;
  }

  /**
   * Initializes the CrowdSec watcher.
   *
   * @return \CrowdSec\CapiClient\Watcher
   *   The CrowdSec watcher.
   */
  public function watcher(): Watcher {
    if (!isset($this->watcher)) {
      $configs = [
        'env' => $this->config->get('env'),
        'api_timeout' => (int) $this->config->get('api_timeout'),
        'scenarios' => $this->config->get('scenarios'),
        'user_agent_suffix' => 'drupal',
        'user_agent_version' => 'v1.0.0',
      ];
      $this->watcher = new Watcher($configs, $this->storage, NULL, $this->logger);
    }
    return $this->watcher;
  }

  /**
   * Initializes the CrowdSec remediation client.
   *
   * @return \CrowdSec\RemediationEngine\CapiRemediation
   *   The remediation client.
   */
  public function remediation(): CapiRemediation {
    if (!isset($this->remediation)) {
      $this->remediation = new CapiRemediation([], $this->watcher(), $this->cache(), $this->logger);
    }
    return $this->remediation;
  }

  /**
   * Refresh data from CrowdSsec.
   */
  public function refresh(): void {
    try {
      $this->remediation()->refreshDecisions();
    }
    catch (InvalidArgumentException | CacheException $e) {
      $this->logger->critical($e->getMessage());
    }
  }

  /**
   * Verifies the IP against the CrowdSec remediation.
   *
   * @param string $ip
   *   The IP address to verify.
   *
   * @return string
   *   The remediation string.
   */
  public function verifyIp(string $ip): string {
    $remediation = $this->remediation();
    try {
      return $remediation->getIpRemediation($ip);
    }
    catch (InvalidArgumentException $e) {
      $this->logger->critical('Unable to receive remediation for @ip: @message', [
        '@ip' => $ip,
        '@message' => $e->getMessage(),
      ]);
    }
    return Constants::REMEDIATION_BYPASS;
  }

  /**
   * Initialize and return the cache backend.
   *
   * @return \CrowdSec\RemediationEngine\CacheStorage\AbstractCache
   *   The initialized cache backend.
   */
  private function cache(): AbstractCache {
    if (!isset($this->cache)) {
      if ($this->config->get('cache_backend_redis') && $redis = Settings::get('redis.connection')) {
        $dns = 'redis://' . ($redis['host'] ?? 'localhost') . ':' . ($redis['port'] ?? '6379');
        $this->cache = new Redis(['redis_dsn' => $dns], $this->logger);
      }
      else {
        $this->cache = new PhpFiles(['fs_cache_path' => $this->fileSystem->getTempDirectory() . '/crowdsec'], $this->logger);
      }

    }
    return $this->cache;
  }

}
