<?php

namespace Drupal\crowdsec\EventSubscriber;

use Drupal\crowdsec\Buffer;
use Drupal\crowdsec\Client;
use Drupal\crowdsec\Form\Settings;
use Drupal\user\Event\UserEvents;
use Drupal\user\Event\UserFloodEvent;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Event\ResponseEvent;
use Symfony\Component\HttpKernel\KernelEvents;

/**
 * CrowdSec event subscriber.
 */
class CrowdSec implements EventSubscriberInterface {

  /**
   * The CrowdSec client service.
   *
   * @var \Drupal\crowdsec\Client
   */
  protected Client $client;

  /**
   * The CrowdSec buffer service.
   *
   * @var \Drupal\crowdsec\Buffer
   */
  protected Buffer $buffer;

  /**
   * Constructs the CrowdSec event subscriber service.
   *
   * @param \Drupal\crowdsec\Client $client
   *   The CrowdSec client service.
   * @param \Drupal\crowdsec\Buffer $buffer
   *   The CrowdSec buffer service.
   */
  public function __construct(Client $client, Buffer $buffer) {
    $this->client = $client;
    $this->buffer = $buffer;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents(): array {
    $events = [];
    $events[KernelEvents::RESPONSE][] = ['onResponse'];
    $events[UserEvents::FLOOD_BLOCKED_IP] = ['onBlockIp'];
    $events[UserEvents::FLOOD_BLOCKED_USER] = ['onBlockIp'];
    return $events;
  }

  /**
   * Will be called when the response was created.
   *
   * @param \Symfony\Component\HttpKernel\Event\ResponseEvent $event
   *   The response event.
   * @param string $eventName
   *   The name of the event.
   */
  public function onResponse(ResponseEvent $event, string $eventName): void {
    $statusCode = $event->getResponse()->getStatusCode();
    if ($statusCode < 400 || $statusCode >= 500) {
      return;
    }
    $this->buffer->addWhisperSignal($event->getRequest()->getClientIp());
  }

  /**
   * Will be called when flood control bans an IP.
   *
   * @param \Drupal\user\Event\UserFloodEvent $event
   *   The user flood event.
   * @param string $eventName
   *   The name of the event.
   */
  public function onBlockIp(UserFloodEvent $event, string $eventName): void {
    $this->buffer->addSignal(Settings::SCENARIO_FLOOD, $event->getIp(), $event->getWindow());
  }

}
