<?php

namespace Drupal\crowdsec\Plugin\QueueWorker;

use Drupal\ban\BanIpManagerInterface;
use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\Core\Queue\QueueInterface;
use Drupal\Core\Queue\QueueWorkerBase;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * CrowdSec queue worker.
 *
 * @QueueWorker(
 *   id = "crowdsec",
 *   title = @Translation("CrowdSec"),
 *   cron = {"time" = 3}
 * )
 */
class CrowdSec extends QueueWorkerBase implements ContainerFactoryPluginInterface {

  /**
   * The logger service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected LoggerChannelInterface $logger;

  /**
   * The queue.
   *
   * @var \Drupal\Core\Queue\QueueInterface
   */
  protected QueueInterface $queue;

  /**
   * The time service.
   *
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected TimeInterface $time;

  /**
   * The ban IP manager.
   *
   * @var \Drupal\ban\BanIpManagerInterface
   */
  protected BanIpManagerInterface $banIpManager;

  /**
   * Constructs the CrowdSec queue worker.
   *
   * @param array $configuration
   *   The plugin configuration.
   * @param string $plugin_id
   *   The plugin ID.
   * @param array $plugin_definition
   *   The plugin definition.
   * @param \Drupal\Core\Logger\LoggerChannelInterface $logger
   *   The logger service.
   * @param \Drupal\Core\Queue\QueueFactory $queueFactory
   *   The queue factory service.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time service.
   * @param \Drupal\ban\BanIpManagerInterface $banIpManager
   *   The ban IP manager.
   */
  public function __construct(array $configuration, string $plugin_id, array $plugin_definition, LoggerChannelInterface $logger, QueueFactory $queueFactory, TimeInterface $time, BanIpManagerInterface $banIpManager) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->logger = $logger;
    $this->queue = $queueFactory->get('crowdsec', TRUE);
    $this->time = $time;
    $this->banIpManager = $banIpManager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): CrowdSec {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('logger.channel.crowdsec'),
      $container->get('queue'),
      $container->get('datetime.time'),
      $container->get('ban.ip_manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function processItem($data): void {
    if ($data['type'] === 'unban' && $data['due'] <= $this->time->getRequestTime()) {
      $this->banIpManager->unbanIp($data['ip']);
      $this->logger->info('Un-banned ip @ip', ['@ip' => $data['ip']]);
      return;
    }

    // Item not processed, let's re-queue.
    $this->queue->createItem($data);
  }

}
